import uuid
from unittest.mock import patch

from django.db.models import Count
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element, TextOrientation, Transcription
from arkindex.images.models import ImageServer
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase


class TestBulkElementTranscriptions(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.rotated_page = Element.objects.create(corpus=cls.corpus, type=cls.page.type, image=cls.page.image, polygon=cls.page.polygon, name="rotated_page", rotation_angle=90)
        cls.mirrored_page = Element.objects.create(corpus=cls.corpus, type=cls.page.type, image=cls.page.image, polygon=cls.page.polygon, name="mirrored_page", mirrored=True)
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.line = cls.corpus.elements.filter(type__slug="text_line").first()
        cls.private_corpus = Corpus.objects.create(name="Private")
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_run = cls.worker_version.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.custom_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.local_worker_run = WorkerRun.objects.get(version=cls.custom_version, process__creator=cls.user)
        cls.huge_image = ImageServer.objects.local.images.create(
            path="kingdom/far/far/away",
            status=S3FileStatus.Checked,
            width=4200,
            height=4200,
        )
        cls.huge_page = Element.objects.create(corpus=cls.corpus, type=cls.page.type, image=cls.huge_image, polygon=[[0, 0], [0, 4200], [4200, 4200], [4200, 0], [0, 0]], name="huge_page")

    def test_bulk_transcriptions_worker_version(self):
        data = {
            "element_type": "text_line",
            "transcriptions": [
                {
                    "polygon": [[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]],
                    "text": "Hello world !",
                    "confidence": 0.1337,
                    "orientation": "vertical-lr",
                }
            ]
        }
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "worker_run_id": ["This field is required."],
        })

    def test_bulk_transcriptions_worker_run(self):
        """
        Bulk creates a list of element with an attached transcription generated by a worker_run
        """
        # Create a manual transcription on the element
        self.line.transcriptions.create(text="A manual transcription")
        self.assertEqual(self.line.transcriptions.count(), 1)

        existing_element_ids = list(Element.objects.get_descending(self.page.id).values_list("id", flat=True))

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337, "vertical-lr"),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], "I <3 JavaScript", 0.42, "horizontal-lr"),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence,
                "orientation": orientation
            } for poly, text, confidence, orientation in transcriptions]
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id__in=existing_element_ids)
        self.assertEqual(created_elts.count(), 2)
        self.assertTrue(all(map(lambda elt: elt.image == self.page.image, created_elts)))
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.polygon.coords, elt.worker_version_id, elt.worker_run_id)
                for elt in created_elts
            ],
            [
                (4, "5", ((13, 37), (13, 137), (133, 137), (133, 37), (13, 37)), self.custom_version.id, self.local_worker_run.id),
                (5, "6", ((24, 42), (24, 142), (64, 142), (64, 42), (24, 42)), self.custom_version.id, self.local_worker_run.id)
            ]
        )
        self.assertCountEqual(
            created_elts.values_list(
                "transcriptions__text",
                "transcriptions__worker_version_id",
                "transcriptions__worker_run_id",
                "transcriptions__orientation"
            ),
            [
                ("Hello world !", self.custom_version.id, self.local_worker_run.id, TextOrientation.VerticalLeftToRight),
                ("I <3 JavaScript", self.custom_version.id, self.local_worker_run.id, TextOrientation.HorizontalLeftToRight)
            ]
        )

    def test_bulk_transcriptions_element_confidence(self):
        """
        Bulk creates a list of elements, with a confidence score and an attached transcription
        """
        # Create a manual transcription on the element
        self.line.transcriptions.create(text="A manual transcription")
        self.assertEqual(self.line.transcriptions.count(), 1)

        existing_element_ids = list(Element.objects.get_descending(self.page.id).values_list("id", flat=True))

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337, "vertical-lr", 0.5),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], "I <3 JavaScript", 0.42, "horizontal-lr", 0.7),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence,
                "orientation": orientation,
                "element_confidence": element_confidence
            } for poly, text, confidence, orientation, element_confidence in transcriptions]
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id__in=existing_element_ids)
        self.assertEqual(created_elts.count(), 2)
        self.assertTrue(all(map(lambda elt: elt.image == self.page.image, created_elts)))
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.polygon.coords, elt.worker_version_id, elt.worker_run_id, elt.confidence)
                for elt in created_elts
            ],
            [
                (4, "5", ((13, 37), (13, 137), (133, 137), (133, 37), (13, 37)), self.custom_version.id, self.local_worker_run.id, 0.5),
                (5, "6", ((24, 42), (24, 142), (64, 142), (64, 42), (24, 42)), self.custom_version.id, self.local_worker_run.id, 0.7)
            ]
        )
        self.assertCountEqual(
            created_elts.values_list(
                "transcriptions__text",
                "transcriptions__worker_version_id",
                "transcriptions__worker_run_id",
                "transcriptions__orientation",
                "transcriptions__confidence"
            ),
            [
                ("Hello world !", self.custom_version.id, self.local_worker_run.id, TextOrientation.VerticalLeftToRight, 0.1337),
                ("I <3 JavaScript", self.custom_version.id, self.local_worker_run.id, TextOrientation.HorizontalLeftToRight, 0.42)
            ]
        )

    @override_settings(ARKINDEX_FEATURES={"search": False})
    def test_bulk_transcriptions_load(self):
        """
        Sending 100 transcriptions should not increase the number of DB queries
        """
        # Create 100 transcriptions
        transcriptions = [(
            [[i, i], [i, i + 20], [i + 20, i + 20], [i + 20, i], [i, i]],
            str(i / 2),
            i / 200
        ) for i in range(1, 200, 2)]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence
            } for poly, text, confidence in transcriptions]
        }

        created_elts = Element.objects.get_descending(self.page.id) \
            .filter(paths__path__last=self.page.id, type__slug="text_line")
        self.assertEqual(created_elts.count(), 1)

        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(created_elts.count(), 101)
        # Each annotated element has a transcription
        self.assertEqual(created_elts.annotate(ts_count=Count("transcriptions")).filter(ts_count=1).count(), 100)

    def test_bulk_transcriptions_similar_zone(self):
        """
        Pushing a transcription matching an element type and zone reuses this element
        Does not erase present transcriptions, even from the same worker
        """
        # Create a transcription from the same worker run on the element
        self.line.transcriptions.create(
            text="A worker run transcription",
            worker_version=self.custom_version,
            worker_run=self.local_worker_run,
        )
        self.assertEqual(self.line.transcriptions.count(), 1)

        existing_element_ids = list(Element.objects.get_descending(self.page.id).values_list("id", flat=True))
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337),
            # Use line zone to create the second transcription
            (list(self.line.polygon), "I <3 JavaScript", 0.42),
            (list(self.line.polygon), "Animula vagula blandula", 0.24),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence
            } for poly, text, confidence in transcriptions]
        }

        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id__in=existing_element_ids)
        # The existing text line has been reused
        self.assertEqual(created_elts.count(), 1)
        self.assertCountEqual(self.line.transcriptions.values_list("text", flat=True), [
            "A worker run transcription",
            "I <3 JavaScript",
            "Animula vagula blandula"
        ])

    def test_bulk_transcriptions_existing_huge_polygon(self):
        polygon = [(i, i) for i in range(3000)]
        polygon.append((0, 0))
        messy_line = self.corpus.elements.create(
            type=self.corpus.types.get(slug="text_line"),
            name=":bazar:",
            image=self.huge_image,
            polygon=polygon,
        )
        messy_line.add_parent(self.huge_page)

        messy_line.transcriptions.create(
            text="Blah",
            worker_version=self.worker_version,
            worker_run=self.local_worker_run,
        )
        self.assertEqual(messy_line.transcriptions.count(), 1)

        existing_element_ids = list(Element.objects.get_descending(self.huge_page.id).values_list("id", flat=True))
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [
                {
                    "polygon": polygon,
                    "text": "Blob",
                    "confidence": 0.42
                }
            ],
        }

        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.huge_page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.huge_page.id).exclude(id__in=existing_element_ids)
        # The existing text line has been reused
        self.assertFalse(created_elts.exists())
        self.assertCountEqual(messy_line.transcriptions.values_list("text", flat=True), [
            "Blah",
            "Blob",
        ])

    def test_bulk_transcriptions_requires_verified(self):
        """
        The user has to be verified tu push transcriptions in bulk
        """
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data={}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_bulk_transcriptions_private(self, filter_rights_mock):
        """
        Accessing a non writeable element triggers a 404_NOT_FOUND
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.private_page.id}),
                data={}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_bulk_transcriptions_non_manual(self):
        """
        Either a worker version or a run are required
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data={
                    "element_type": "text_line",
                    "transcriptions": [{
                        "polygon": [[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]],
                        "text": "Can I write womething here ?",
                        "confidence": .666
                    }]
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": ["This field is required."],
        })

    def test_bulk_transcriptions_wrong_fields_validation(self):
        """
        Briefly assert fields values are validated
        """
        bad_id = str(uuid.uuid4())
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data={
                    "element_type": "paragraph",
                    "worker_run_id": bad_id,
                    "transcriptions": [
                        {
                            "polygon": [[13, 37]],
                            "text": "There is a snake in my computer",
                            "confidence": 0.00001,
                        },
                        {
                            "polygon": [[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]],
                            "text": "No.",
                        },
                        {
                            "polygon": [[13, 37], [133, 37], [-133, 137], [13, 137], [13, 37]],
                            "text": "Thinking outside the box",
                        }
                    ]
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "element_type": ["Object with slug=paragraph does not exist."],
            "worker_run_id": [f'Invalid pk "{bad_id}" - object does not exist.'],
            "transcriptions": [
                {
                    "polygon": ["Ensure this field has at least 3 elements."]
                },
                {
                    "confidence": ["This field is required."],
                },
                {
                    "confidence": ["This field is required."],
                    "polygon": {
                        "2": {
                            "0": ["Ensure this value is greater than or equal to 0."],
                        }
                    }
                },
            ]
        })

    @override_settings(ARKINDEX_FEATURES={"search": False})
    def test_bulk_transcriptions_return_ordered_elements(self):
        """
        return_elements field allows to serialize an ordered list of sub-elements holding the transcriptions
        """
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "stuck", 0.13),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], "stock", 0.37),
            # The existing line has the smallest ordering
            (self.line.polygon.coords, "stack", 0.42),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence
            } for poly, text, confidence in transcriptions],
            "return_elements": True
        }

        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id=self.line.id)
        self.assertEqual(created_elts.count(), 5)

        first_elt = created_elts.get(transcriptions__text="stuck")
        second_elt = created_elts.get(transcriptions__text="stock")
        first_tr = Transcription.objects.get(text="stuck")
        second_tr = Transcription.objects.get(text="stock")
        third_td = Transcription.objects.get(text="stack")
        self.assertListEqual(response.json(), [{
            "id": str(first_tr.id),
            "element_id": str(first_elt.id),
            "created": True
        }, {
            "id": str(second_tr.id),
            "element_id": str(second_elt.id),
            "created": True
        }, {
            "id": str(third_td.id),
            "element_id": str(self.line.id),
            "created": False
        }])

    def test_bulk_transcriptions_requires_zone(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.vol.id}),
                data={}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"element": ["Cannot create transcriptions on an element without a zone."]})

    def test_top_level_element(self):
        """
        Create transcriptions on a top level element
        """
        # Create a top level page
        top_level = self.corpus.elements.create(
            type=self.page.type,
            name="Top level page",
            image=self.page.image,
            polygon=self.page.polygon,
        )

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], "I <3 JavaScript", 0.42),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence
            } for poly, text, confidence in transcriptions]
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": top_level.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(top_level.id)
        self.assertEqual(created_elts.count(), 2)
        self.assertTrue(all(map(lambda elt: elt.image == self.page.image, created_elts)))
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.polygon.coords, elt.worker_version.id, elt.worker_run_id)
                for elt in created_elts
            ],
            [
                (0, "1", ((13, 37), (13, 137), (133, 137), (133, 37), (13, 37)), self.custom_version.id, self.local_worker_run.id),
                (1, "2", ((24, 42), (24, 142), (64, 142), (64, 42), (24, 42)), self.custom_version.id, self.local_worker_run.id),
            ]
        )
        self.assertCountEqual(
            created_elts.values_list("transcriptions__text", "transcriptions__worker_version_id", "transcriptions__worker_run_id"),
            [
                ("Hello world !", self.custom_version.id, self.local_worker_run.id),
                ("I <3 JavaScript", self.custom_version.id, self.local_worker_run.id),
            ]
        )

    def test_bulk_transcriptions_default_orientation(self):
        """
        No text orientation set > default (horizontal-lr) value
        """
        existing_element_ids = list(Element.objects.get_descending(self.page.id).values_list("id", flat=True))
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], "I <3 JavaScript", 0.42),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence,
            } for poly, text, confidence in transcriptions]
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_elts = Element.objects.get_descending(self.page.id).exclude(id__in=existing_element_ids)
        self.assertCountEqual(
            created_elts.values_list("transcriptions__orientation", flat=True),
            [TextOrientation.HorizontalLeftToRight, TextOrientation.HorizontalLeftToRight]
        )

    def test_bulk_transcriptions_invalid_orientation(self):
        """
        Specifying an invalid text orientation throws an error
        """
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337, "wiggly"),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], "I <3 JavaScript", 0.42, "timey-wimey"),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence,
                "orientation": orientation,
            } for poly, text, confidence, orientation in transcriptions]
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"transcriptions": [
            {"orientation": ["Value is not of type TextOrientation"]},
            {"orientation": ["Value is not of type TextOrientation"]}
        ]}
        )

    def test_bulk_transcriptions_inherit_rotation(self):
        """
        The created subelements inherit their parent's rotation angle
        """
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], "I <3 JavaScript", 0.42),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence,
            } for poly, text, confidence in transcriptions]
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.rotated_page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_elts = Element.objects.get_descending(self.rotated_page.id)
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.polygon.coords, elt.worker_version.id, elt.worker_run_id, elt.rotation_angle)
                for elt in created_elts
            ],
            [
                (0, "1", ((13.0, 37.0), (13.0, 137.0), (133.0, 137.0), (133.0, 37.0), (13.0, 37.0)), self.custom_version.id, self.local_worker_run.id, 90),
                (1, "2", ((24.0, 42.0), (24.0, 142.0), (64.0, 142.0), (64.0, 42.0), (24.0, 42.0)), self.custom_version.id, self.local_worker_run.id, 90),
            ]
        )

    def test_bulk_transcriptions_inherit_mirroring(self):
        """
        The created subelements are mirrored if their parent is
        """
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], "I <3 JavaScript", 0.42),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence,
            } for poly, text, confidence in transcriptions]
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.mirrored_page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_elts = Element.objects.get_descending(self.mirrored_page.id)
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.polygon.coords, elt.worker_version.id, elt.worker_run_id, elt.mirrored)
                for elt in created_elts
            ],
            [
                (0, "1", ((13.0, 37.0), (13.0, 137.0), (133.0, 137.0), (133.0, 37.0), (13.0, 37.0)), self.custom_version.id, self.local_worker_run.id, True),
                (1, "2", ((24.0, 42.0), (24.0, 142.0), (64.0, 142.0), (64.0, 42.0), (24.0, 42.0)), self.custom_version.id, self.local_worker_run.id, True)
            ]
        )

    def test_bulk_transcriptions_outside_image(self):
        """
        Cannot create transcribed elements with a polygon with one or more points outside the parent image
        """
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], "Hello world !", 0.1337),
            ([[24, 42], [64, 42], [64, 142], [2400, 142], [24, 42]], "I <3 JavaScript", 0.42),
        ]
        data = {
            "element_type": "text_line",
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [{
                "polygon": poly,
                "text": text,
                "confidence": confidence,
            } for poly, text, confidence in transcriptions]
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "transcriptions": {"1": {"polygon": ["A transcription element's polygon must not exceed its image's bounds."]}}
        })

    def test_worker_run_non_local(self):
        """
        A regular user cannot create transcribed elements with a WorkerRun of a non-local process
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data={
                    "parent": str(self.page.id),
                    "element_type": "text_line",
                    "transcriptions": [
                        {
                            "polygon": [[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]],
                            "text": "I <3 JavaScript",
                            "confidence": .42
                        },
                    ],
                    "worker_run_id": str(self.worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_run_id": [
            "Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."
        ]})

    def test_worker_run_other_user(self):
        """
        A regular user cannot create transcribed elements with a WorkerRun of someone else's local process
        """
        worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data={
                    "parent": str(self.page.id),
                    "element_type": "text_line",
                    "transcriptions": [
                        {
                            "polygon": [[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]],
                            "text": "I <3 JavaScript",
                            "confidence": .42
                        },
                    ],
                    "worker_run_id": str(worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_run_id": [
            "Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."
        ]})

    def test_worker_run_other_process(self):
        """
        A Ponos task cannot create transcribed elements with a WorkerRun of another process
        """
        process2 = self.worker_run.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_run.version)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data={
                    "parent": str(self.page.id),
                    "element_type": "text_line",
                    "transcriptions": [
                        {
                            "polygon": [[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]],
                            "text": "I <3 JavaScript",
                            "confidence": .42
                        },
                    ],
                    "worker_run_id": str(other_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Only the WorkerRuns of the authenticated task's process may be used."
            ]
        })

    def test_create_task_auth(self):
        """
        Transcribed elements can be created with a WorkerRun of a non-local process
        when authenticated as a Ponos task of this process
        """
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data={
                    "parent": str(self.page.id),
                    "element_type": "text_line",
                    "transcriptions": [
                        {
                            "polygon": [[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]],
                            "text": "I <3 JavaScript",
                            "confidence": .42
                        },
                    ],
                    "worker_run_id": str(self.worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.worker_version_id, list(elt.transcriptions.values_list("text", flat=True)))
                for elt in Element.objects.get_descending(self.page.id).filter(
                    type__slug="text_line",
                    worker_run=self.worker_run,
                )
            ],
            [
                (4, "5", self.worker_run.version.id, ["I <3 JavaScript"]),
            ]
        )

    def test_worker_run_local_task_auth(self):
        """
        Transcribed elements can be created with a WorkerRun of a local process
        even when authenticated as a Ponos task from a different process
        """
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:element-transcriptions-bulk", kwargs={"pk": self.page.id}),
                data={
                    "parent": str(self.page.id),
                    "element_type": "text_line",
                    "transcriptions": [
                        {
                            "polygon": [[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]],
                            "text": "I <3 JavaScript",
                            "confidence": .42
                        },
                    ],
                    "worker_run_id": str(self.local_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.worker_version_id, list(elt.transcriptions.values_list("text", flat=True)))
                for elt in Element.objects.get_descending(self.page.id).filter(
                    type__slug="text_line",
                    worker_run=self.local_worker_run,
                )
            ],
            [
                (4, "5", self.local_worker_run.version.id, ["I <3 JavaScript"]),
            ]
        )
