from unittest.mock import call, patch

from django.conf import settings
from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Farm
from arkindex.process.models import ArkindexFeature, DataFile, Process, ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestImportFromFiles(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.image_datafile = cls.corpus.files.create(
            name="test.jpg",
            size=42,
            content_type="image/jpg",
        )
        cls.pdf_datafile = cls.corpus.files.create(
            name="test.pdf",
            size=42,
            content_type="application/pdf",
        )
        cls.iiif_datafile = cls.corpus.files.create(
            name="test.json",
            size=42,
            content_type="application/json",
        )

    def test_from_files_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(self.image_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(23):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(self.image_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.image_datafile])
        self.assertEqual(process.folder_type.slug, "volume")
        self.assertEqual(process.element_type.slug, "page")
        self.assertEqual(process.farm, None)
        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport))
        import_task = process.tasks.get(slug="import_files")
        self.assertEqual(
            import_task.env["ARKINDEX_WORKER_RUN_ID"],
            str(worker_run.id),
        )

    def test_from_files_pdf(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(23):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(self.pdf_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.pdf_datafile])
        self.assertIsNone(process.element)

    def test_from_files_iiif(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(23):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(self.iiif_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.iiif_datafile])
        self.assertIsNone(process.element)

    def test_from_files_multiple_types(self):
        self.client.force_login(self.user)

        archive_datafiles = DataFile.objects.bulk_create(
            DataFile(
                corpus=self.corpus,
                name=content_type,
                content_type=content_type,
                size=1337,
            )
            for content_type in settings.ARCHIVE_MIME_TYPES
        )

        with self.assertNumQueries(36):
            response = self.client.post(reverse("api:files-process"), {
                "files": [
                    str(self.pdf_datafile.id),
                    str(self.image_datafile.id),
                    str(self.iiif_datafile.id),
                    *(datafile.id for datafile in archive_datafiles),
                ],
                "mode": "files",
                "folder_type": "volume",
                "element_type": "page",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertQuerySetEqual(
            process.files.all(), [
                self.image_datafile,
                self.pdf_datafile,
                self.iiif_datafile,
                *archive_datafiles,
            ],
            ordered=False,
        )
        self.assertIsNone(process.element)

    def test_from_files_files_wrong_type(self):
        self.client.force_login(self.user)
        bad_datafile = self.corpus.files.create(
            name="test.mp4",
            size=42,
            content_type="video/mp4",
        )

        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(bad_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"non_field_errors": ["File imports can only import images, PDF documents, ZIP archives or IIIF manifests"]})

    def test_from_files_mode_ignored(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(23):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(self.iiif_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
                "mode": "easy"
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.mode, ProcessMode.Files)

    def test_from_files_folder_id(self):
        self.client.force_login(self.user)
        volume = self.corpus.elements.get(name="Volume 1")
        with self.assertNumQueries(26):
            response = self.client.post(
                reverse("api:files-process"),
                {
                    "files": [str(self.pdf_datafile.id)],
                    "mode": "files",
                    "folder_id": str(volume.id),
                    "folder_type": "volume",
                    "element_type": "page",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.element, volume)

    def test_from_files_folder_type_required(self):
        self.client.force_login(self.user)
        volume = self.corpus.elements.get(name="Volume 1")
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:files-process"),
                {
                    "files": [str(self.pdf_datafile.id)],
                    "mode": "files",
                    "folder_id": str(volume.id),
                    "element_type": "page",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"folder_type": ["This field is required."]})

    def test_from_files_no_files(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:files-process"), {
                "files": [],
                "folder_type": "volume",
                "element_type": "page",
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_farm(self):
        self.client.force_login(self.user)
        farm = Farm.objects.get(name="Wheat farm")

        with self.assertNumQueries(24):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(self.image_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
                "farm_id": str(farm.id),
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.farm, farm)

    @patch("arkindex.process.serializers.imports.get_default_farm")
    def test_from_files_default_farm(self, get_default_farm_mock):
        farm = Farm.objects.get(name="Wheat farm")
        get_default_farm_mock.return_value = farm
        self.client.force_login(self.user)

        with self.assertNumQueries(23):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(self.image_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data["id"])
        self.assertEqual(process.farm, farm)

    @patch("arkindex.ponos.models.Farm.is_available", return_value=False)
    def test_from_files_farm_guest(self, is_available_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:files-process"), {
                "files": [str(self.image_datafile.id)],
                "folder_type": "volume",
                "element_type": "page",
                "farm_id": str(Farm.objects.first().id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "farm_id": ["You do not have access to this farm."],
        })
        self.assertEqual(is_available_mock.call_count, 1)
        self.assertEqual(is_available_mock.call_args, call(self.user))
